static const char *RcsId = "$Header: /home/cvsadm/cvsroot/fermi/servers/hdb++/hdb++es/src/PushThread.cpp,v 1.7 2014-03-06 15:21:43 graziano Exp $";
//+=============================================================================
//
// file :         PushThread.cpp
//
// description :  C++ source for the PushThread
//
//
//
//-=============================================================================

#include "PushThread.h"
#include "AlarmNotify.h"

namespace AlarmNotify_ns
{
    //=============================================================================
    //=============================================================================
    PushThread::PushThread(Tango::DeviceImpl *dev,
            const string &ds_name, const vector<string>& configuration): AbortableThread(dev)
                                                                                         , new_data(&new_data_mutex)
                                                                                         , max_waiting(0)
    {
        _device = dev;
        unordered_map<string,string> mail_conf;
        AlarmStringUtils::string_vector2map(configuration, "=", mail_conf);
        try
        {
            dp = std::unique_ptr<Tango::DeviceProxy>(new Tango::DeviceProxy(mail_conf.at("device")));
        }
        catch(const std::out_of_range& oe)
        {
            stringstream tmp;
            tmp << "Configuration parsing error looking for device: " << oe.what();
            DEBUG_STREAM << __func__<< ": " << tmp.str() << endl;
            dp = std::unique_ptr<Tango::DeviceProxy>(nullptr);
        }
        catch(Tango::DevFailed& fe)
        {
            stringstream tmp;
            tmp << "Error creating Proxy: " << fe.errors[0].desc;
            DEBUG_STREAM << __func__<< ": " << tmp.str() << endl;
            dp = std::unique_ptr<Tango::DeviceProxy>(nullptr);
        }
        try
        {
            command = mail_conf.at("command");
        }
        catch(const std::out_of_range& oe)
        {
            stringstream tmp;
            tmp << "Configuration parsing error looking for command: " << oe.what();
            DEBUG_STREAM << __func__<< ": " << tmp.str() << endl;
        }      

    }

    //=============================================================================
    //=============================================================================
    void PushThread::push_back_cmd(std::unique_ptr<AlarmData> argin)
    {
        AlarmNotify *_dev = static_cast<AlarmNotify *>(_device);
        //omni_mutex_lock sync(new_data_mutex);
        //	Add data at end of vector
        if(!is_aborted())
        {
            size_t events_size = 0; 
            {
                omni_mutex_lock lock(new_data_mutex);
                events.push_back(std::move(argin));
                events_size = events.size();
                //	Check if nb waiting more the stored one.
                if (events_size > max_waiting)
                    max_waiting = events_size;
                
                _dev->AttributePendingNumber = events_size;
                _dev->AttributeMaxPendingNumber = max_waiting;
            }
        }
        // Signal only if there was no data
        new_data.signal();
    }

    //=============================================================================
    //=============================================================================
    auto PushThread::nb_cmd_waiting() -> size_t
    {
        omni_mutex_lock sync(new_data_mutex);
        return events.size();
    }
    //=============================================================================
    //=============================================================================
    auto PushThread::get_max_waiting() -> size_t
    {
        omni_mutex_lock sync(new_data_mutex);
        return max_waiting;
    }
    //=============================================================================
    //=============================================================================
    void PushThread::get_sig_list_waiting(vector<string>& list)
    {
        omni_mutex_lock sync(new_data_mutex);
        list.clear();
        for (auto& ev : events)
        {
            string signame;
            signame = string(ev->ev_data->attr_name);
            //signame = string(ev->signal.name);
            list.push_back(signame);
        }
    }

    //=============================================================================
    //=============================================================================
    auto PushThread::get_next_cmds() -> std::vector<std::unique_ptr<AlarmData>>
    {
        AlarmNotify *_dev = static_cast<AlarmNotify *>(_device);
        size_t events_size = 0; 
        
        omni_mutex_lock sync(new_data_mutex);
            
        events_size = events.size();

        while(events_size == 0 && !is_aborted())
        {
            new_data.wait();
            events_size = events.size();
        }

        std::vector<std::unique_ptr<AlarmData>> cmds;
        for(size_t i = 0; i < events_size; ++i)
        {
            cmds.push_back(std::move(events.front()));
            events.pop_front();
        }
        
        // We cleared the list, so this should be 0.
        _dev->AttributePendingNumber = 0;

        return cmds;
    }

    void PushThread::init_abort_loop()
    {
    }

    //=============================================================================
    /**
     * Execute the thread infinite loop.
     */
    //=============================================================================
    void PushThread::run_thread_loop()
    {
        AlarmNotify *_dev = static_cast<AlarmNotify *>(_device);
        //	Check if command ready
        std::vector<std::unique_ptr<AlarmData>> cmds;
        while (!(cmds = get_next_cmds()).empty())
        {
            std::vector<std::tuple<Tango::EventData *, AlarmEventDataType>> events;
            std::vector<std::tuple<AlarmSignal&, std::chrono::duration<double>>> signals;
            bool batch = batch_insert && cmds.size() > 1;
            for(const auto& cmd : cmds)
            {
                std::chrono::duration<double> rcv_time = std::chrono::seconds(cmd->ev_data->get_date().tv_sec) + std::chrono::milliseconds(cmd->ev_data->get_date().tv_usec);
                if(batch)
                {
                    events.emplace_back(std::make_tuple(cmd->ev_data, cmd->ev_data_type));
                    
                    signals.push_back({cmd->signal, rcv_time});
                }
                else
                {
                    /*for(auto iii : cmd->alarm_info)
                    {
                        DEBUG_STREAM << __func__ << ": key=" << iii.first << " value=" << iii.second;
                    }*/
                    if(dp == nullptr)
                    {
                        INFO_STREAM << __func__ << ": Cannot send email for " << cmd->ev_data->attr_name << " because Proxy not connected";
                        continue;
                    }
#if 0 //TODO: need also to retrieve enum labels
                    Tango::DevEnum val_r;
                    try
                    {
                        cmd->ev_data->attr_value->extract_read(val_r);
                    }
                    catch(Tango::DevFailed &e)
                    {

                    }
#endif
                    string groups;
                    try
                    {
                        string tag, handler;
                        string attr_name =  AlarmStringUtils::get_only_attr_name(cmd->ev_data->attr_name);
                        AlarmStringUtils::split_tag_handler(attr_name, handler, tag);

                        DEBUG_STREAM << __func__ << ": pushing event " << cmd->ev_data->attr_name << " into email";
                        ostringstream email;
                        string state;//TODO: get state from event value!!
                        email << "name=" << cmd->alarm_info.at("tag") << ";";
                        email << "handler=" << handler << ";";
                        email << "values=" << cmd->alarm_info.at("values") << ";";
                        email << "url=" << cmd->alarm_info.at("url") << ";";
                        email << "msg=" << cmd->alarm_info.at("message") << ";";
                        email << "formula=" << cmd->alarm_info.at("formula") << ";";
                        groups = cmd->alarm_info.at("group");
                        email << "groups=" << groups << ";";
                        email << "shlvd_time=" << cmd->alarm_info.at("shlvd_time") << ";";
                        state = cmd->alarm_info.at("state");//TODO: get state from event value!!
                        email << "state=" << state << ";";
                        email << "ack=" << ((cmd->alarm_info.at("ack") == "0") ? "false" : "true") << ";";
                        email << "time=" << cmd->alarm_info.at("time") << ";";
                        DEBUG_STREAM << __func__ << ": ... -> state=" << state << " name=" << cmd->alarm_info.at("tag");
                        cmd->signal.set_alarm_state(state);

                        string receivers, prev_receivers;
                        string prev_state;
                        try
                        {
                            prev_state = _dev->prevStateMap.at(cmd->ev_data->attr_name);
                        }
                        catch(const std::out_of_range& oe)
                        {
                            DEBUG_STREAM << __func__<< ": No previous state for '"<<  cmd->ev_data->attr_name << "'" << endl;
                        }
                        
                        _dev->prepare_receivers(cmd->ev_data->attr_name, state, groups, receivers);
                        if(state!= prev_state && (prev_state == "OOSRV" || prev_state == "SHLVD")) //to allow notification from disabled
                        {
                            _dev->prepare_receivers(cmd->ev_data->attr_name, prev_state, groups, prev_receivers);
                            if(!prev_receivers.empty())
                                DEBUG_STREAM << __func__ << ": ADDED previous receiver alarm='" << cmd->ev_data->attr_name << "' state=" << state << " prev_state=" << prev_state << " rec=" << prev_receivers;
                        }
                        if(!receivers.empty() && !prev_receivers.empty())
                            receivers = receivers + "," +  prev_receivers;
                        else if(receivers.empty() && !prev_receivers.empty())
                            receivers = prev_receivers;

                        if(!receivers.empty())
                        {
                            email << "dest=";
                            email << receivers;
                            email << ";";
                            DEBUG_STREAM << __func__ << ": ...-> " << email.str();
                            Tango::DeviceData Din;
                            Din << email.str().c_str();
                            dp->command_inout(command,Din);
                            DEBUG_STREAM << __func__ << ": sent email for " << cmd->ev_data->attr_name << ": -->";
                            DEBUG_STREAM << __func__ << ":  --> " << email.str();                    
                            cmd->signal.set_ok_db();
                            _dev->prevStateMap[cmd->ev_data->attr_name] = state;
                        }
                        
#if 0
                        Tango::DeviceData Din;
                        Din << email.str().c_str();
                        if(cmd->alarm_info.at("state") == "UNACK" || cmd->alarm_info.at("state") == "ACKED")
                            dp->command_inout("SendAlarm",Din);
                        else if(cmd->alarm_info.at("state") == "NORM" || cmd->alarm_info.at("state") == "RTNUN")
                            dp->command_inout("SendNormal",Din);
                        else //TODO: ERROR, OOSRV, SHLVD
                            dp->command_inout("SendAlarm",Din);
                        //mdb->insert_event(cmd->ev_data, cmd->ev_data_type);//TODO:!!!!!!!!!!!!
                        DEBUG_STREAM << __func__ << ": sent email for " << cmd->ev_data->attr_name;
#endif
                    }
                    catch(const std::out_of_range& oe)
                    {
                        cmd->signal.set_nok_db(string(oe.what()));
                        stringstream tmp;
                        tmp << "Alarm info parsing error looking for tag: " << oe.what();
                        DEBUG_STREAM << __func__<< ": " << tmp.str() << endl;
                    }
                    catch(Tango::DevFailed  &e)
                    {
                        cmd->signal.set_nok_db(string(e.errors[0].desc));
                        INFO_STREAM << __func__ << ": ERROR sending email for " << cmd->ev_data->attr_name;
                        Tango::Except::print_exception(e);
                    }
                }
            }
            if(!events.empty())
            {
                try
                {
                    DEBUG_STREAM << __func__ << ": HERE pushing event batch size=" << events.size();
                    //mdb->insert_events(events);//TODO:!!!!!!!!!!!!

                    // We can't get the individual speed for each signal
                    for(const auto& sig : signals)
                    {
                        std::get<0>(sig).set_ok_db();
                    }
                }
                catch(Tango::DevFailed  &e)
                {
                    for(size_t i = 0; i < events.size(); ++i)
                    {
                        AlarmSignal& signal = std::get<0>(signals[i]);
                        try
                        {
                            DEBUG_STREAM << __func__ << ": HERE falling back from batch to push event " << std::get<0>(events[i])->attr_name;
                            //mdb->insert_event(std::get<0>(events[i]), std::get<1>(events[i]));//TODO:!!!!!!!!!!!!

                            signal.set_ok_db();
                        }
                        catch(Tango::DevFailed  &e)
                        {
                            signal.set_nok_db(string(e.errors[0].desc));
                            Tango::Except::print_exception(e);
                        }
                    }
                }
            }
        }
    }

    void PushThread::finalize_abort_loop()
    {
        cout <<"PushThread::"<< __func__<<": exiting..."<<endl;
    }

    auto PushThread::get_abort_loop_period() -> std::chrono::milliseconds
    {
        //using namespace std::chrono_literals;
        //return -1s;TODO:!!!
        return std::chrono::milliseconds(-1000);
    }

    void PushThread::do_abort()
    {
        // if waiting for new cmd, signal.
        new_data.signal();
    }
}//	namespace
